# generated by datamodel-codegen:
#   filename:  ecs-files-input.json

from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, List, Optional, Union

S3Uri = str


ComposeXUri = str


class Encoding(str, Enum):
    base64 = "base64"
    plain = "plain"


class Context(str, Enum):
    plain = "plain"
    jinja2 = "jinja2"


@dataclass
class IgnoreFailureItem:
    commands: Optional[bool] = False
    mode: Optional[bool] = False
    owner: Optional[bool] = False
    source_download: Optional[bool] = False


@dataclass
class UrlDef:
    Url: Optional[str] = None
    Username: Optional[str] = None
    Password: Optional[str] = None


@dataclass
class IamOverrideDef:
    """
    When source points to AWS, allows to indicate if another role should be used
    """

    RoleArn: Optional[str] = None
    SessionName: Optional[str] = "S3File@EcsConfigComposer"
    ExternalId: Optional[str] = None
    RegionName: Optional[str] = None
    AccessKeyId: Optional[str] = None
    SecretAccessKey: Optional[str] = None
    SessionToken: Optional[str] = None


@dataclass
class CommandsDefItem:
    """
    Command to run with options
    """

    command: Optional[str] = None
    display_output: Optional[bool] = False
    ignore_error: Optional[bool] = False


CommandsDef = List[Union[str, CommandsDefItem]]


@dataclass
class X509CertDef:
    keyFileName: str
    certFileName: str
    dir_path: Optional[str] = None
    emailAddress: Optional[str] = "files-composer@compose-x.tld"
    commonName: Optional[str] = None
    countryName: Optional[str] = "ZZ"
    localityName: Optional[str] = "Anywhere"
    stateOrProvinceName: Optional[str] = "Shire"
    organizationName: Optional[str] = "NoOne"
    organizationUnitName: Optional[str] = "Automation"
    validityEndInSeconds: Optional[float] = 8035200
    group: Optional[str] = "root"
    owner: Optional[str] = "root"


@dataclass
class Certificates:
    x509: Optional[Dict[str, X509CertDef]] = None


@dataclass
class Commands:
    post: Optional[CommandsDef] = None
    pre: Optional[CommandsDef] = None


@dataclass
class SsmDef:
    ParameterName: Optional[str] = None
    IamOverride: Optional[IamOverrideDef] = None


@dataclass
class SecretDef:
    SecretId: str
    VersionId: Optional[str] = None
    VersionStage: Optional[str] = None
    JsonKey: Optional[str] = None
    IamOverride: Optional[IamOverrideDef] = None


@dataclass
class S3Def:
    S3Uri: Optional[S3Uri] = None
    ComposeXUri: Optional[ComposeXUri] = None
    BucketName: Optional[str] = None
    BucketRegion: Optional[str] = None
    Key: Optional[str] = None
    IamOverride: Optional[IamOverrideDef] = None


@dataclass
class SourceDef:
    Url: Optional[UrlDef] = None
    Ssm: Optional[SsmDef] = None
    S3: Optional[S3Def] = None
    Secret: Optional[SecretDef] = None


@dataclass
class FileDef:
    path: Optional[str] = None
    content: Optional[str] = None
    source: Optional[SourceDef] = None
    encoding: Optional[Encoding] = "plain"
    group: Optional[str] = "root"
    owner: Optional[str] = "root"
    mode: Optional[str] = "0644"
    context: Optional[Context] = "plain"
    ignore_failure: Optional[Union[IgnoreFailureItem, bool]] = None
    commands: Optional[Commands] = None


@dataclass
class Model:
    """
    Configuration input definition for ECS Files Composer
    """

    files: Optional[Dict[str, FileDef]] = None
    certificates: Optional[Certificates] = None
    IamOverride: Optional[IamOverrideDef] = None



# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2022 John Mille<john@compose-x.io>

"""
AWS Based Functions
"""

from __future__ import annotations

import json
import re
from os import environ

import requests
from aws_cfn_custom_resource_resolve_parser import handle
from boto3.session import Session
from botocore.exceptions import ClientError
from compose_x_common.aws.ec2 import get_ec2_subnet_from_vpc_and_ip_cidr
from compose_x_common.compose_x_common import keyisset

from ecs_files_composer.jinja2_filters import get_property


def define_ecs_metadata(for_task=False):
    meta_v4 = "ECS_CONTAINER_METADATA_URI_V4"
    meta_v3 = "ECS_CONTAINER_METADATA_URI"

    if environ.get(meta_v4, None):
        meta_url = environ.get(meta_v4)
    elif environ.get(meta_v3, None):
        meta_url = environ.get(meta_v3)
    else:
        raise OSError("No ECS Metadata URL provided. This filter only works on ECS")
    if for_task:
        return requests.get(f"{meta_url}/task")
    else:
        return requests.get(meta_url)


def msk_bootstrap(msk_arn: str, broker_type: str) -> str:
    """
    Uses the ARN of a MSK cluster,
    and returns the list of BootStrap endpoints for a private MSK cluster using SASL IAM.
    If failed, returns the ARN.
    """
    session = Session()
    client = session.client("kafka")
    brokers_r = client.get_bootstrap_brokers(ClusterArn=msk_arn)
    if keyisset(broker_type, brokers_r):
        return brokers_r[broker_type]
    return msk_arn


def msk_cluster_zookeeper(msk_arn, with_tls: bool):
    session = Session()
    client = session.client("kafka")
    config_r = client.describe_cluster(ClusterArn=msk_arn)
    config_info = config_r["ClusterInfo"]
    if with_tls and keyisset("ZookeeperConnectStringTls", config_info):
        return config_info["ZookeeperConnectStringTls"]
    else:
        if keyisset("ZookeeperConnectString", config_info):
            return config_info["ZookeeperConnectString"]
    return ""


def msk_endpoints(msk_arn: str, broker_type: str, endpoint_type: str):
    session = Session()
    client = session.client("kafka")
    brokers_r = client.get_bootstrap_brokers(ClusterArn=msk_arn)
    if not keyisset(broker_type, brokers_r):
        return
    _endpoints = brokers_r[broker_type].split(",")
    endpoint_port: dict = {"JMX": 11001, "NODE": 11002}
    _msk_ports_re = re.compile(r":(?P<port>\d+)")
    if endpoint_type not in endpoint_port:
        endpoint_type = "JMX"
    endpoints: list[str] = []
    for _endpoint in _endpoints:
        endpoints.append(
            _msk_ports_re.sub(f":{endpoint_port[endpoint_type]}", _endpoint)
        )
    return endpoints


def from_ssm_json(parameter_name: str) -> dict:
    """
    Function to retrieve an SSM parameter value

    :param parameter_name: Name of the parameter
    """
    value_str = (
        Session()
        .client("ssm")
        .get_parameter(Name=parameter_name, WithDecryption=True)["Parameter"]["Value"]
    )
    try:
        return json.loads(value_str)
    except json.JSONDecodeError:
        return {}


def from_ssm(parameter_name: str) -> str:
    """
    Function to retrieve an SSM parameter value

    :param parameter_name: Name of the parameter
    """
    return (
        Session()
        .client("ssm")
        .get_parameter(Name=parameter_name, WithDecryption=True)["Parameter"]["Value"]
    )


def ecs_container_metadata(property_key=None, fallback_value=None):
    metadata_raw = define_ecs_metadata()
    metadata = metadata_raw.json()
    if property_key:
        value = get_property(metadata, property_key)
        if value is None:
            print(f"No container property found matching {property_key}")
            return fallback_value
        return value
    return metadata


def ecs_task_metadata(property_key=None, fallback_value=None):
    metadata_raw = define_ecs_metadata(for_task=True)
    metadata = metadata_raw.json()
    if property_key:
        value = get_property(metadata, property_key)
        if value is None:
            print(f"No task property found matching {property_key}")
            return fallback_value
        return value
    return metadata


def using_resolve(resolve_string: str) -> str:
    return handle(resolve_string)


def ec2_zone_id(subnet_id: str = None):
    """
    Defines which AWS ZoneID the container is into based on the subnet if provided, otherwise using EC2 Region API
    """
    session = Session()
    if not subnet_id:
        vpc_id = ecs_task_metadata("VPCID")
        subnet_range = ecs_container_metadata("Networks::0::IPv4SubnetCIDRBlock")
        subnet_details = get_ec2_subnet_from_vpc_and_ip_cidr(
            subnet_range, vpc_id, session
        )
        return subnet_details["AvailabilityZoneId"]
    elif subnet_id:
        ec2_client = session.client("ec2")
        try:
            subnet_r = ec2_client.describe_subnets(SubnetIds=[subnet_id])
            return subnet_r["Subnets"][0]["AvailabilityZoneId"]
        except ClientError as error:
            print("Failed to retrieve ZoneID from Subnet API call", error)
    print("Unable to find subnet using API calls")
    return ""


def dump_ecs_details():
    print("Dumping ECS Details")
    try:
        print("ECS Container Metadata: ", ecs_container_metadata())
        print("ECS Task Metadata: ", ecs_task_metadata())
        vpc_id = ecs_task_metadata("VPCID")
        subnet_range = ecs_container_metadata("Networks::0::IPv4SubnetCIDRBlock")
        print("VPC ID: ", vpc_id)
        print("Subnet Range: ", subnet_range)
        try:
            subnet_details = get_ec2_subnet_from_vpc_and_ip_cidr(
                subnet_range, vpc_id, Session()
            )
            print("Container subnet details: ", subnet_details)
        except Exception as error:
            print("Could not retrieve subnet details", error)
    except Exception as error:
        print(error)



# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2022 John Mille<john@compose-x.io>

"""Main module."""

from __future__ import annotations

import json
import os
from dataclasses import asdict
from enum import Enum
from os import environ, path
from tempfile import TemporaryDirectory
from typing import TYPE_CHECKING, ByteString

import yaml
from dacite import Config, from_dict
from yaml import Loader

from ecs_files_composer import input
from ecs_files_composer.aws_mgmt import S3Fetcher
from ecs_files_composer.certificates_mgmt import process_x509_certs
from ecs_files_composer.common import LOG
from ecs_files_composer.files_mgmt import File


def init_config(
    raw=None,
    file_path=None,
    env_var=None,
    ssm_parameter=None,
    s3_config=None,
    secret_config=None,
    role_arn=None,
    external_id=None,
    decode_base64=False,
    context=None,
    override_folder: str = None,
    print_generated_config: bool = False,
):
    """Function to initialize the configuration as if it were a file itself"""
    iam_override = {"SessionName": "FilesComposerInit"}
    if ssm_parameter or s3_config or secret_config:
        role_arn = environ.get("CONFIG_IAM_ROLE_ARN", role_arn)
        external_id = environ.get("CONFIG_IAM_EXTERNAL_ID", external_id)
        if role_arn:
            iam_override.update({"RoleArn": role_arn})
        if external_id:
            iam_override.update({"ExternalId": external_id})
    if ssm_parameter:
        initial_config = {"source": {"Ssm": {"ParameterName": ssm_parameter}}}
    elif s3_config:
        if not S3Fetcher.bucket_re.match(s3_config):
            raise ValueError(
                "The value for S3 URI is not valid.",
                s3_config,
                "Expected to match",
                S3Fetcher.bucket_re.pattern,
            )
        initial_config = {
            "source": {
                "S3": {
                    "BucketName": S3Fetcher.bucket_re.match(s3_config).group("bucket"),
                    "Key": S3Fetcher.bucket_re.match(s3_config).group("key"),
                }
            }
        }
    elif secret_config:
        initial_config = {"source": {"Secret": {"SecretId": secret_config}}}
    elif file_path:
        with open(path.abspath(file_path)) as file_fd:
            config_content = file_fd.read()
        initial_config = {"content": config_content}
    elif raw:
        initial_config = {"content": raw}
    elif env_var:
        LOG.debug(f"Using env var {env_var}")
        initial_config = {"content": environ.get(env_var, None)}
    else:
        raise ValueError("No input source was provided")
    if not initial_config:
        raise ImportError("Failed to import a configuration content")
    LOG.debug(initial_config)
    if not override_folder:
        temp_dir = TemporaryDirectory()
        config_path = f"{temp_dir.name}/init.conf"
    else:
        config_path = f"{override_folder}/init.conf"
    jobs_input_def = {
        "files": {config_path: initial_config},
        "IamOverride": iam_override,
    }
    if decode_base64:
        initial_config["encoding"] = "base64"
    if context:
        initial_config["context"] = context
    start_jobs(jobs_input_def)
    try:
        with open(config_path) as config_fd:
            _file_content = config_fd.read()
            if print_generated_config:
                LOG.info(_file_content)
            try:
                config = yaml.load(_file_content, Loader=Loader)
                LOG.info(f"Successfully loaded YAML config {config_path}")
            except yaml.YAMLError:
                try:
                    config = json.loads(_file_content)
                    LOG.info(f"Successfully loaded JSON config {config_path}")
                except json.JSONDecodeError:
                    LOG.error("Input content is not valid JSON")
                    raise
            finally:
                if print_generated_config:
                    LOG.info(jobs_input_def)
            return config

    except OSError as error:
        LOG.exception(error)
        LOG.error(f"Failed to read input file from {config_path}")


def process_files(job: input.Model, override_session=None) -> None:
    files: list = []
    for file_path, file in job.files.items():
        if not isinstance(file, File):
            file_redef = from_dict(
                data_class=File, data=asdict(file), config=Config(cast=[Enum, bytes])
            )
            file_redef.path = file_path
            files.append(file_redef)
        else:
            files.append(file)
    for file in files:
        file.handler(job.IamOverride, override_session)
        LOG.info(f"Tasks for {file.path} completed.")


def start_jobs(config: dict, override_session=None):
    """Starting point to run the files job"""
    job = from_dict(
        data_class=input.Model, data=config, config=Config(cast=[Enum, bytes])
    )
    if job.certificates:
        process_x509_certs(job)
    if job.files:
        process_files(job, override_session)



# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2022 John Mille<john@compose-x.io>

"""Top-level package for ECS Files Composer."""

__author__ = """John Preston"""
__email__ = "john@compose-x.io"
__version__ = "2.0.0"



# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2022 John Mille<john@compose-x.io>

from __future__ import annotations

from os import environ

from ecs_files_composer.jinja2_functions.aws import (
    ec2_zone_id,
    ecs_container_metadata,
    ecs_task_metadata,
    from_ssm,
    from_ssm_json,
    msk_bootstrap,
    using_resolve,
)


def env_var(key, value=None):
    return environ.get(key, value)


def hostname(alternative_value: str = None) -> str:
    try:
        import platform

        return str(platform.node())
    except Exception as error:
        print("Error with platform", error)
        try:
            import socket

            return str(socket.gethostname())
        except Exception as error:
            print("Error with socket", error)
    if alternative_value:
        return alternative_value


JINJA_FUNCTIONS = {
    "ecs_container_metadata": ecs_container_metadata,
    "ecs_task_metadata": ecs_task_metadata,
    "env_var": env_var,
    "from_ssm": from_ssm,
    "from_ssm_json": from_ssm_json,
    "from_resolve": using_resolve,
    "msk_bootstrap": msk_bootstrap,
    "hostname": hostname,
    "ec2_zone_id": ec2_zone_id,
    "subnet_zone_id": ec2_zone_id,
}



# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2022 John Mille<john@compose-x.io>

"""Console script for ecs_files_composer."""
import argparse
import sys
from os import environ

from ecs_files_composer.common import LOG
from ecs_files_composer.ecs_files_composer import init_config, start_jobs
from ecs_files_composer.jinja2_functions.aws import dump_ecs_details


def main():
    """Console script for ecs_files_composer."""
    parser = argparse.ArgumentParser()
    options = parser.add_mutually_exclusive_group()
    options.add_argument(
        "-f",
        "--from-file",
        help="Configuration for execution from a file",
        type=str,
        required=False,
        dest="file_path",
    )
    options.add_argument(
        "-e",
        "--from-env-var",
        dest="env_var",
        required=False,
        help="Configuration for execution is in an environment variable",
    )
    options.add_argument(
        "--from-ssm",
        dest="ssm_config",
        help="Configuration for execution is in an SSM Parameter",
        required=False,
    )
    options.add_argument(
        "--from-s3",
        dest="s3_config",
        required=False,
        help="Configuration for execution is in an S3",
    )
    options.add_argument(
        "--from-secrets",
        dest="secret_config",
        required=False,
        help="Configuration for execution is in an AWS Secrets Manager",
    )
    parser.add_argument(
        "--role-arn",
        help="The Role ARN to use for the configuration initialization",
        required=False,
    )
    parser.add_argument(
        "--decode-base64",
        help="Whether the source config is in base64 encoded",
        action="store_true",
        required=False,
        default=False,
    )
    parser.add_argument(
        "--context",
        help="Indicate which context to use (valid: jinja2|plain). Default is jinja2",
        required=False,
        default="jinja2",
    )
    parser.add_argument(
        "--override-init-folder",
        dest="init_folder",
        required=False,
        type=str,
        default="",
    )
    parser.add_argument(
        "--dump-ecs-details",
        action="store_true",
        required=False,
        default=False,
        help="On startup, dumps ecs metadata to stdout",
    )
    parser.add_argument(
        "--print-generated-config",
        action="store_true",
        help="Display generated config",
        default=False,
    )
    args = parser.parse_args()
    LOG.debug(f"CLI ARGS?: {args}")
    if args.dump_ecs_details:
        dump_ecs_details()
    if not (
        args.env_var or args.ssm_config or args.s3_config or args.file_path
    ) and environ.get("ECS_CONFIG_CONTENT", None):
        LOG.info("Using default env variable ECS_CONFIG_CONTENT")
        config = init_config(
            env_var="ECS_CONFIG_CONTENT",
            decode_base64=bool(environ.get("DECODE_BASE64", False)),
            context=environ.get("context", "jinja2"),
            print_generated_config=args.print_generated_config,
        )
    elif args.env_var:
        config = init_config(
            env_var=args.env_var,
            decode_base64=args.decode_base64,
            context=args.context,
            override_folder=args.init_folder,
            print_generated_config=args.print_generated_config,
        )
    elif args.file_path:
        config = init_config(
            file_path=args.file_path,
            decode_base64=args.decode_base64,
            context=args.context,
            override_folder=args.init_folder,
            print_generated_config=args.print_generated_config,
        )
    elif args.ssm_config:
        config = init_config(
            ssm_parameter=args.ssm_config,
            decode_base64=args.decode_base64,
            context=args.context,
            override_folder=args.init_folder,
            print_generated_config=args.print_generated_config,
        )
    elif args.s3_config:
        config = init_config(
            s3_config=args.s3_config,
            decode_base64=args.decode_base64,
            context=args.context,
            override_folder=args.init_folder,
            print_generated_config=args.print_generated_config,
        )
    elif args.secret_config:
        config = init_config(
            secret_config=args.secret_config,
            decode_base64=args.decode_base64,
            context=args.context,
            override_folder=args.init_folder,
            print_generated_config=args.print_generated_config,
        )
    else:
        raise parser.error(
            "You must specify where the execution configuration comes from or set ECS_CONFIG_CONTENT."
        )
    start_jobs(config)
    return 0


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2022 John Mille<john@compose-x.io>

"""
Package allowing to expand the Jinja filters to use.
"""
import json
import re
from base64 import b64decode, b64encode
from os import environ

import yaml
from flatdict import FlatterDict

from ecs_files_composer.common import LOG


def env_override(value, key):
    """
    Function to use in new Jinja filter
    :param value:
    :param key:
    :return:
    """
    return environ.get(key, value)


def from_list_to_dict(top_key, new_mapping, to_convert):
    """

    :param str top_key:
    :param dict new_mapping:
    :param list to_convert:
    :return:
    """
    for count, item in enumerate(to_convert):
        list_key = f"{top_key}_{count}"
        if isinstance(item, dict):
            from_dict_to_simple_keys(list_key, new_mapping, item)
        elif isinstance(item, list):
            from_list_to_dict(list_key, new_mapping, item)


def from_dict_to_simple_keys(top_key, new_mapping, to_convert):
    """

    :param str top_key:
    :param dict new_mapping:
    :param dict to_convert:
    :return:
    """
    for key, value in to_convert.items():
        if isinstance(value, str):
            new_mapping[f"{top_key}_{key}"] = value
        elif isinstance(value, dict):
            from_dict_to_simple_keys(f"{top_key}_{key}", new_mapping, value)
        elif isinstance(value, list):
            from_list_to_dict(f"{top_key}_{key}", new_mapping, value)


def from_metadata_to_flat_keys(metadata):
    """
    Function to transform the metadata into simplified structure
    :param dict metadata:
    :return:
    """
    new_metadata = {}
    for key, value in metadata.items():
        if isinstance(value, str):
            new_metadata[key] = value
        elif isinstance(value, dict):
            from_dict_to_simple_keys(key, new_metadata, value)
        elif isinstance(value, list):
            from_list_to_dict(key, new_metadata, value)
    return new_metadata


def get_property(metadata, property_key, separator: str = None):
    if separator is None:
        separator = r"::"
    metadata_mapping = FlatterDict(metadata)
    metadata_mapping.set_delimiter(separator)
    LOG.debug(metadata_mapping)
    if property_key in metadata_mapping:
        return metadata_mapping[property_key]

    metadata_mapping = from_metadata_to_flat_keys(metadata)
    LOG.debug(metadata_mapping)
    property_re = re.compile(property_key)
    for key, value in metadata_mapping.items():
        if property_re.findall(key):
            return value
    return None


def to_yaml(value):
    """
    Filter to render input to YAML formatted content
    :return:
    """
    return yaml.dump(value, Dumper=yaml.Dumper)


def to_json(value, indent=2):
    return json.dumps(value, indent=indent)


def base64encode(value: str):
    """Return value base64 encoded"""
    try:
        return b64encode(value).decode("utf-8")
    except TypeError:
        return b64encode(value.encode("utf-8")).decode("utf-8")


def base64decode(value) -> bytes:
    """Decodes base64 encoded value"""
    return b64decode(value)


JINJA_FILTERS = {
    "to_yaml": to_yaml,
    "to_json": to_json,
    "env_override": env_override,
    "base64encode": base64encode,
    "base64decode": base64decode,
}
